# pylint:disable=missing-class-docstring
import itertools
from collections import defaultdict
from typing import Union, Type, Callable

import networkx

from .typevars import (
    Existence,
    Equivalence,
    Subtype,
    TypeVariable,
    DerivedTypeVariable,
    HasField,
    Add,
    ConvertTo,
    IsArray,
)
from .typeconsts import (
    BottomType,
    TopType,
    TypeConstant,
    Int,
    Int8,
    Int16,
    Int32,
    Int64,
    Pointer,
    Pointer32,
    Pointer64,
    Struct,
    int_type,
    TypeVariableReference,
)

# lattice for 64-bit binaries
BASE_LATTICE_64 = networkx.DiGraph()
BASE_LATTICE_64.add_edge(TopType, Int)
BASE_LATTICE_64.add_edge(Int, Int64)
BASE_LATTICE_64.add_edge(Int, Int32)
BASE_LATTICE_64.add_edge(Int, Int16)
BASE_LATTICE_64.add_edge(Int, Int8)
BASE_LATTICE_64.add_edge(Int32, BottomType)
BASE_LATTICE_64.add_edge(Int16, BottomType)
BASE_LATTICE_64.add_edge(Int8, BottomType)
BASE_LATTICE_64.add_edge(Int64, Pointer64)
BASE_LATTICE_64.add_edge(Pointer64, BottomType)

# lattice for 32-bit binaries
BASE_LATTICE_32 = networkx.DiGraph()
BASE_LATTICE_32.add_edge(TopType, Int)
BASE_LATTICE_32.add_edge(Int, Int64)
BASE_LATTICE_32.add_edge(Int, Int32)
BASE_LATTICE_32.add_edge(Int, Int16)
BASE_LATTICE_32.add_edge(Int, Int8)
BASE_LATTICE_32.add_edge(Int32, Pointer32)
BASE_LATTICE_32.add_edge(Int64, BottomType)
BASE_LATTICE_32.add_edge(Pointer32, BottomType)
BASE_LATTICE_32.add_edge(Int16, BottomType)
BASE_LATTICE_32.add_edge(Int8, BottomType)

BASE_LATTICES = {
    32: BASE_LATTICE_32,
    64: BASE_LATTICE_64,
}


class RecursiveType:
    def __init__(self, typevar, offset):
        self.typevar = typevar
        self.offset = offset


class SimpleSolver:
    """
    SimpleSolver is, literally, a simple, unification-based type constraint solver.
    """

    def __init__(self, bits: int, constraints):
        if bits not in (32, 64):
            raise ValueError("Pointer size %d is not supported. Expect 32 or 64." % bits)

        self.bits = bits
        self._constraints = constraints
        self._base_lattice = BASE_LATTICES[bits]

        #
        # Solving state
        #
        self._equivalence = {}
        self._lower_bounds = defaultdict(BottomType)
        self._upper_bounds = defaultdict(TopType)
        self._recursive_types = defaultdict(set)

        self.solve()
        self.solution = self.determine()

    def solve(self):
        # import pprint
        # pprint.pprint(self._constraints)

        eq_constraints = self._eq_constraints_from_add()
        self._constraints |= eq_constraints
        constraints = self._handle_equivalence()
        subtypevars, supertypevars = self._calculate_closure(constraints)
        self._find_recursive_types(subtypevars)
        self._compute_lower_upper_bounds(subtypevars, supertypevars)
        self._lower_struct_fields()
        self._convert_arrays(constraints)
        # import pprint
        # print("Lower bounds")
        # pprint.pprint(self._lower_bounds)
        # print("Upper bounds")
        # pprint.pprint(self._upper_bounds)

    def determine(self):
        solution = {}

        for v in self._lower_bounds:
            if isinstance(v, TypeVariable) and not isinstance(v, DerivedTypeVariable):
                lb = self._lower_bounds[v]
                if isinstance(lb, BottomType):
                    # use its upper bound instead
                    solution[v] = self._upper_bounds[v]
                else:
                    solution[v] = lb

        for v in self._upper_bounds:
            if v not in solution:
                ub = self._upper_bounds[v]
                if not isinstance(ub, TopType):
                    solution[v] = ub

        for v, e in self._equivalence.items():
            if v not in solution:
                solution[v] = solution.get(e, None)

        # import pprint
        # print("Lower bounds")
        # pprint.pprint(self._lower_bounds)
        # print("Upper bounds")
        # pprint.pprint(self._upper_bounds)
        # print("Solution")
        # pprint.pprint(solution)
        return solution

    def _handle_equivalence(self):
        graph = networkx.Graph()

        replacements = {}
        constraints = set()

        # collect equivalence relations
        for constraint in self._constraints:
            if isinstance(constraint, Equivalence):
                # | type_a == type_b
                # we apply unification and removes one of them
                ta, tb = constraint.type_a, constraint.type_b
                if isinstance(ta, TypeConstant) and isinstance(tb, TypeVariable):
                    # replace tb with ta
                    replacements[tb] = ta
                elif isinstance(ta, TypeVariable) and isinstance(tb, TypeConstant):
                    # replace ta with tb
                    replacements[ta] = tb
                else:
                    # they are both type variables. we will determine a representative later
                    graph.add_edge(ta, tb)

        for components in networkx.connected_components(graph):
            components_lst = list(sorted(components, key=lambda x: str(x)))  # pylint:disable=unnecessary-lambda
            representative = components_lst[0]
            for tv in components_lst[1:]:
                replacements[tv] = representative

        # replace
        for constraint in self._constraints:
            if isinstance(constraint, Existence):
                replaced, new_constraint = constraint.replace(replacements)

                if replaced:
                    constraints.add(new_constraint)
                else:
                    constraints.add(constraint)

            elif isinstance(constraint, Subtype):
                # subtype <: supertype
                # replace type variables
                replaced, new_constraint = constraint.replace(replacements)

                if replaced:
                    constraints.add(new_constraint)
                else:
                    constraints.add(constraint)

        # import pprint
        # print("Replacements")
        # pprint.pprint(replacements)
        # print("Constraints (after replacement)")
        # pprint.pprint(constraints)

        self._equivalence = replacements
        return constraints

    def _eq_constraints_from_add(self):
        """
        Handle Add constraints.
        """
        new_constraints = set()
        for constraint in self._constraints:
            if isinstance(constraint, Add):
                if (
                    isinstance(constraint.type_0, TypeVariable)
                    and not isinstance(constraint.type_0, DerivedTypeVariable)
                    and isinstance(constraint.type_r, TypeVariable)
                    and not isinstance(constraint.type_r, DerivedTypeVariable)
                ):
                    new_constraints.add(Equivalence(constraint.type_0, constraint.type_r))
                if (
                    isinstance(constraint.type_1, TypeVariable)
                    and not isinstance(constraint.type_1, DerivedTypeVariable)
                    and isinstance(constraint.type_r, TypeVariable)
                    and not isinstance(constraint.type_r, DerivedTypeVariable)
                ):
                    new_constraints.add(Equivalence(constraint.type_1, constraint.type_r))
        return new_constraints

    def _pointer_class(self) -> Union[Type[Pointer32], Type[Pointer64]]:
        if self.bits == 32:
            return Pointer32
        elif self.bits == 64:
            return Pointer64
        raise NotImplementedError("Unsupported bits %d" % self.bits)

    def _calculate_closure(self, constraints):
        ptr_class = self._pointer_class()

        # a mapping from type variables to all the variables which are {super,sub}types of them
        subtypevars = defaultdict(set)  # {k: {v}}: v <: k
        supertypevars = defaultdict(set)  # {k: {v}}: k <: v

        constraints = set(constraints)  # make a copy

        while constraints:
            constraint = constraints.pop()

            if isinstance(constraint, Existence):
                # has a derived type
                if isinstance(constraint.type_, DerivedTypeVariable):
                    # handle label
                    if isinstance(constraint.type_.label, HasField):
                        # the original variable is a pointer
                        v = constraint.type_.type_var.type_var
                        if isinstance(v, TypeVariable):
                            subtypevars[v].add(
                                ptr_class(
                                    Struct(
                                        fields={
                                            constraint.type_.label.offset: int_type(constraint.type_.label.bits),
                                        }
                                    )
                                )
                            )

            elif isinstance(constraint, Subtype):
                # subtype <: supertype

                subtype, supertype = constraint.sub_type, constraint.super_type

                if isinstance(supertype, TypeVariable):
                    if subtype not in subtypevars[supertype]:
                        if supertype is not subtype:
                            subtypevars[supertype].add(subtype)
                            for s in supertypevars[subtype]:
                                # re-add impacted constraints
                                constraints.add(Subtype(subtype, s))

                    if subtype in subtypevars:
                        for v in subtypevars[subtype]:
                            if v not in subtypevars[supertype]:
                                if supertype is not v:
                                    subtypevars[supertype].add(v)
                                    for sup in supertypevars[v]:
                                        constraints.add(Subtype(subtype, sup))

                if isinstance(subtype, TypeVariable):
                    if supertype not in supertypevars[subtype]:
                        if subtype is not supertype:
                            supertypevars[subtype].add(supertype)
                            for s in subtypevars[supertype]:
                                # re-add impacted constraints
                                constraints.add(Subtype(s, supertype))

                    if supertype in supertypevars:
                        for v in supertypevars[supertype]:
                            if v not in supertypevars[subtype]:
                                if v is not subtype:
                                    supertypevars[subtype].add(v)
                                    for sup in supertypevars[v]:
                                        constraints.add(Subtype(subtype, sup))

            elif isinstance(constraint, Equivalence):
                raise Exception("Shouldn't exist anymore.")

            else:
                raise NotImplementedError("Unsupported instance type %s." % type(constraint))

        # import pprint
        # print("Subtype vars")
        # pprint.pprint(subtypevars)
        # print("Supertype vars")
        # pprint.pprint(supertypevars)

        return subtypevars, supertypevars

    def _find_recursive_types(self, subtypevars):
        ptr_class = self._pointer_class()

        for var in list(subtypevars.keys()):
            sts = subtypevars[var].copy()
            if isinstance(var, DerivedTypeVariable) and isinstance(var.label, HasField):
                for subtype_var in sts:
                    if var.type_var.type_var == subtype_var:
                        subtypevars[subtype_var].add(
                            ptr_class(Struct({var.label.offset: TypeVariableReference(subtype_var)}))
                        )
                        self._recursive_types[subtype_var].add(var.label.offset)

    def _get_lower_bound(self, v):
        if isinstance(v, TypeConstant):
            return v
        return self._lower_bounds[v]

    def _get_upper_bound(self, v):
        if isinstance(v, TypeConstant):
            return v
        if v in self._upper_bounds:
            return self._upper_bounds[v]

        # try to compute it
        if isinstance(v, DerivedTypeVariable):
            if isinstance(v.label, ConvertTo):
                # after integer conversion,
                ub = int_type(v.label.to_bits)
                if ub is not None:
                    self._upper_bounds[v] = ub
            elif isinstance(v.label, HasField):
                ub = int_type(v.label.bits)
                if ub is not None:
                    self._upper_bounds[v] = ub

        # if all that failed, let the defaultdict generate a Top
        return self._upper_bounds[v]

    def _compute_lower_upper_bounds(self, subtypevars, supertypevars):
        # compute the least upper bound for each type variable
        for typevar, upper_bounds in supertypevars.items():
            if typevar is None:
                continue
            if isinstance(typevar, TypeConstant):
                continue
            self._upper_bounds[typevar] = self._meet(typevar, *upper_bounds, translate=self._get_upper_bound)

        # compute the greatest lower bound for each type variable
        seen = set()  # loop avoidance
        queue = list(subtypevars)
        while queue:
            typevar = queue.pop(0)
            lower_bounds = subtypevars[typevar]

            if typevar not in seen:
                # we detect if it depends on any other typevar upon the first encounter
                seen.add(typevar)

                abort = False
                for subtypevar in lower_bounds:
                    if isinstance(subtypevar, TypeVariable) and subtypevar not in self._lower_bounds:
                        # oops - we should analyze the subtypevar first
                        queue.append(typevar)
                        # to avoid loops, make sure typevar does not rely on
                        abort = True
                        break
                if abort:
                    continue
            else:
                # avoid loop and continue no matter what
                pass

            self._lower_bounds[typevar] = self._join(typevar, *lower_bounds, translate=self._get_lower_bound)

            # because of T-InheritR, fields are propagated *both ways* in a subtype relation
            for subtypevar in lower_bounds:
                if not isinstance(subtypevar, TypeVariable):
                    continue
                subtype_infimum = self._lower_bounds[subtypevar]
                if isinstance(subtype_infimum, Pointer) and isinstance(subtype_infimum.basetype, Struct):
                    subtype_infimum = self._join(subtypevar, typevar, translate=self._get_lower_bound)
                    self._lower_bounds[subtypevar] = subtype_infimum

    def _lower_struct_fields(self):
        # tv_680: ptr32(struct{0: int32})
        # tv_680.load.<32>@0: ptr32(struct{5: int8})
        #    becomes
        # tv_680: ptr32(struct{0: ptr32(struct{5: int8})})

        for outer, outer_lb in self._lower_bounds.items():
            if (
                isinstance(outer, DerivedTypeVariable)
                and isinstance(outer.label, HasField)
                and not isinstance(outer_lb, BottomType)
            ):
                # unpack v
                base = outer.type_var.type_var

                if base in self._lower_bounds:
                    base_lb = self._lower_bounds[base]

                    # make sure it's a pointer at the offset that v.label specifies
                    if isinstance(base_lb, Pointer):
                        if isinstance(base_lb.basetype, Struct):
                            the_field = base_lb.basetype.fields[outer.label.offset]
                            # replace this field
                            new_field = self._meet(the_field, outer_lb, translate=self._get_upper_bound)
                            if new_field != the_field:
                                new_fields = base_lb.basetype.fields.copy()
                                new_fields.update(
                                    {
                                        outer.label.offset: new_field,
                                    }
                                )
                                base_lb = base_lb.__class__(Struct(new_fields))
                                self._lower_bounds[base] = base_lb

                            # another attempt: if a pointer to a struct has only one field, remove the struct
                            if len(base_lb.basetype.fields) == 1 and 0 in base_lb.basetype.fields:
                                base_lb = base_lb.__class__(base_lb.basetype.fields[0])
                                self._lower_bounds[base] = base_lb

    def _convert_arrays(self, constraints):
        for constraint in constraints:
            if not isinstance(constraint, Existence):
                continue
            inner = constraint.type_
            if isinstance(inner, DerivedTypeVariable) and isinstance(inner.label, IsArray):
                if inner.type_var in self._lower_bounds:
                    curr_type = self._lower_bounds[inner.type_var]
                    if isinstance(curr_type, Pointer) and isinstance(curr_type.basetype, Struct):
                        # replace all fields with the first field
                        if 0 in curr_type.basetype.fields:
                            first_field = curr_type.basetype.fields[0]
                            for offset in curr_type.basetype.fields.keys():
                                curr_type.basetype.fields[offset] = first_field

    def _abstract(self, t):  # pylint:disable=no-self-use
        return t.__class__

    def _concretize(self, n_cls, t1, t2, join_or_meet, translate):
        ptr_class = self._pointer_class()

        if n_cls is ptr_class:
            if isinstance(t1, ptr_class) and isinstance(t2, ptr_class):
                # we need to merge them
                return ptr_class(join_or_meet(t1.basetype, t2.basetype, translate=translate))
            if isinstance(t1, ptr_class):
                return t1
            elif isinstance(t2, ptr_class):
                return t2
            else:
                # huh?
                return ptr_class(BottomType())

        return n_cls()

    def _join(self, *args, translate: Callable):
        """
        Get the least upper bound (V, maximum) of the arguments.
        """

        if len(args) == 0:
            return BottomType()
        if len(args) == 1:
            return translate(args[0])
        if len(args) > 2:
            split = len(args) // 2
            first = self._join(*args[:split], translate=translate)
            second = self._join(*args[split:], translate=translate)
            return self._join(first, second, translate=translate)

        t1 = translate(args[0])
        t2 = translate(args[1])

        # Trivial cases
        if t1 == t2:
            return t1
        if isinstance(t1, TopType):
            return t1
        elif isinstance(t2, TopType):
            return t2
        if isinstance(t1, BottomType):
            return t2
        elif isinstance(t2, BottomType):
            return t1
        if isinstance(t1, TypeVariableReference) and not isinstance(t2, TypeVariableReference):
            return t1
        elif isinstance(t2, TypeVariableReference) and not isinstance(t1, TypeVariableReference):
            return t2

        # consult the graph
        t1_cls = self._abstract(t1)
        t2_cls = self._abstract(t2)

        if t1_cls in self._base_lattice and t2_cls in self._base_lattice:
            queue = [t1_cls]
            while queue:
                n = queue[0]
                queue = queue[1:]

                if networkx.has_path(self._base_lattice, n, t2_cls):
                    return self._concretize(n, t1, t2, self._join, translate)
                # go up
                queue.extend(self._base_lattice.predecessors(n))

        # handling Struct
        if t1_cls is Struct and t2_cls is Struct:
            fields = {}
            for offset in sorted(set(itertools.chain(t1.fields.keys(), t2.fields.keys()))):
                if offset in t1.fields and offset in t2.fields:
                    v = self._join(t1.fields[offset], t2.fields[offset], translate=translate)
                elif offset in t1.fields:
                    v = t1.fields[offset]
                elif offset in t2.fields:
                    v = t2.fields[offset]
                else:
                    raise Exception("Impossible")
                fields[offset] = v
            return Struct(fields=fields)

        # single element and single-element struct
        if issubclass(t2_cls, Int) and t1_cls is Struct:
            # swap them
            t1, t1_cls, t2, t2_cls = t2, t2_cls, t1, t1_cls
        if issubclass(t1_cls, Int) and t2_cls is Struct and len(t2.fields) == 1 and 0 in t2.fields:
            # e.g., char & struct {0: char}
            return Struct(fields={0: self._join(t1, t2.fields[0], translate=translate)})

        ptr_class = self._pointer_class()

        # Struct and Pointers
        if t1_cls is ptr_class and t2_cls is Struct:
            # swap them
            t1, t1_cls, t2, t2_cls = t2, t2_cls, t1, t1_cls
        if t1_cls is Struct and len(t1.fields) == 1 and 0 in t1.fields:
            if t1.fields[0].size == 8 and t2_cls is Pointer64:
                # they are equivalent
                # e.g., struct{0: int64}  ptr64(int8)
                # return t2 since t2 is more specific
                return t2
            elif t1.fields[0].size == 4 and t2_cls is Pointer32:
                return t2

        # import ipdb; ipdb.set_trace()
        return TopType()

    def _meet(self, *args, translate: Callable):
        """
        Get the greatest lower bound (^, minimum) of the arguments.
        """

        if len(args) == 0:
            return TopType()
        if len(args) == 1:
            return translate(args[0])
        if len(args) > 2:
            split = len(args) // 2
            first = self._meet(*args[:split], translate=translate)
            second = self._meet(*args[split:], translate=translate)
            return self._meet(first, second, translate=translate)

        t1 = translate(args[0])
        t2 = translate(args[1])

        # Trivial cases
        if t1 == t2:
            return t1
        elif isinstance(t1, BottomType):
            return t1
        elif isinstance(t2, BottomType):
            return t2
        if isinstance(t1, TopType):
            return t2
        elif isinstance(t2, TopType):
            return t1
        if isinstance(t1, TypeVariableReference) and not isinstance(t2, TypeVariableReference):
            return t1
        elif isinstance(t2, TypeVariableReference) and not isinstance(t1, TypeVariableReference):
            return t2

        # consult the graph
        t1_cls = self._abstract(t1)
        t2_cls = self._abstract(t2)

        if t1_cls in self._base_lattice and t2_cls in self._base_lattice:
            queue = [t1_cls]
            while queue:
                n = queue[0]
                queue = queue[1:]

                if networkx.has_path(self._base_lattice, t2_cls, n):
                    return self._concretize(n, t1, t2, self._meet, translate)
                # go down
                queue.extend(self._base_lattice.successors(n))

        # handling Struct
        if t1_cls is Struct and t2_cls is Struct:
            fields = {}
            for offset in sorted(set(itertools.chain(t1.fields.keys(), t2.fields.keys()))):
                if offset in t1.fields and offset in t2.fields:
                    v = self._meet(t1.fields[offset], t2.fields[offset], translate=translate)
                elif offset in t1.fields:
                    v = t1.fields[offset]
                elif offset in t2.fields:
                    v = t2.fields[offset]
                else:
                    raise Exception("Impossible")
                fields[offset] = v
            return Struct(fields=fields)

        # single element and single-element struct
        if issubclass(t2_cls, Int) and t1_cls is Struct:
            # swap them
            t1, t1_cls, t2, t2_cls = t2, t2_cls, t1, t1_cls
        if issubclass(t1_cls, Int) and t2_cls is Struct and len(t2.fields) == 1 and 0 in t2.fields:
            # e.g., char & struct {0: char}
            return Struct(fields={0: self._meet(t1, t2.fields[0], translate=translate)})

        ptr_class = self._pointer_class()

        # Struct and Pointers
        if t1_cls is ptr_class and t2_cls is Struct:
            # swap them
            t1, t1_cls, t2, t2_cls = t2, t2_cls, t1, t1_cls
        if t1_cls is Struct and len(t1.fields) == 1 and 0 in t1.fields:
            if t1.fields[0].size == 8 and t2_cls is Pointer64:
                # they are equivalent
                # e.g., struct{0: int64}  ptr64(int8)
                # return t2 since t2 is more specific
                return t2
            elif t1.fields[0].size == 4 and t2_cls is Pointer32:
                return t2

        # import ipdb; ipdb.set_trace()
        return BottomType()



#!/usr/bin/env python3
# pylint: disable=missing-class-docstring,no-self-use
__package__ = __package__ or "tests.sim"  # pylint:disable=redefined-builtin

import gc
import os
import pickle
import unittest

import angr
from angr.state_plugins.history import HistoryIter

from ..common import bin_location, slow_test


test_location = os.path.join(bin_location, "tests")

target_addrs = {
    "i386": [0x080485C9],
    "x86_64": [0x4006ED],
    "ppc": [0x1000060C],
    "armel": [0x85F0],
    "android/arm": [0x4004CC],
    "mips": [0x4009FC],
}

avoid_addrs = {
    "i386": [0x080485DD, 0x08048564],
    "x86_64": [0x4006AA, 0x4006FD],
    "ppc": [0x10000644, 0x1000059C],
    "armel": [0x86F8, 0x857C],
    "android/arm": [0x4004F0, 0x400470],
    "mips": [0x400A10, 0x400774],
}

corrupt_addrs = {
    "i386": [0x80486B6, b"bO\xcc", lambda s: s.memory.store(s.regs.esp, s.regs.eax)],
    "x86_64": [0x400742, b"\x0f\x0b\xb0[\x41", lambda s: s.registers.store("rdx", 8)],
    "ppc": [0x100006B8, b"\x05\xad\xc2\xea", lambda s: s.registers.store("r5", 8)],
    "armel": [0x8678, b"\xbdM\xec3", lambda s: s.registers.store("r2", 8)],
    "mips": [0x400918, b"[\xf8\x96@"[::-1], lambda s: s.registers.store("a2", 8)],
}

divergences = {
    "ppc": 0x10000588,
    "x86_64": 0x40068E,
    "i386": 0x8048559,
    "armel": 0x8568,
    "android/arm": 0x40045C,
    "mips": 0x40075C,
}


class TestFauxware(unittest.TestCase):
    def _run_fauxware(self, arch):
        p = angr.Project(os.path.join(test_location, arch, "fauxware"), auto_load_libs=False)
        results = p.factory.simulation_manager().explore(find=target_addrs[arch], avoid=avoid_addrs[arch])
        stdin = results.found[0].posix.dumps(0)
        assert b"\x00\x00\x00\x00\x00\x00\x00\x00\x00SOSNEAKY\x00" == stdin

        # test the divergence detection
        ancestor = results.found[0].history.closest_common_ancestor((results.avoid + results.active)[0].history)
        divergent_point = list(HistoryIter(results.found[0].history, end=ancestor))[0]
        # p.factory.block(divergent_point.addr).pp()
        assert divergent_point.recent_bbl_addrs[0] == divergences[arch]

    def _run_pickling(self, arch):
        p = angr.Project(os.path.join(test_location, arch, "fauxware"), auto_load_libs=False)
        pg = p.factory.simulation_manager().run(n=10)
        pickled = pickle.dumps(pg, pickle.HIGHEST_PROTOCOL)
        del p
        del pg
        gc.collect()
        pg = pickle.loads(pickled)

        pg.explore(find=target_addrs[arch], avoid=avoid_addrs[arch])
        stdin = pg.found[0].posix.dumps(0)
        assert b"\x00\x00\x00\x00\x00\x00\x00\x00\x00SOSNEAKY\x00" == stdin

    @slow_test
    def _run_fastmem(self, arch):
        p = angr.Project(os.path.join(test_location, arch, "fauxware"), auto_load_libs=False)
        p.analyses.CongruencyCheck(throw=True).set_state_options(right_add_options={"FAST_REGISTERS"}).run()

    def _run_nodecode(self, arch):
        p = angr.Project(os.path.join(test_location, arch, "fauxware"), auto_load_libs=False)

        # screw up the instructions and make sure the test fails with nodecode
        for i, c in enumerate(corrupt_addrs[arch][1]):
            p.loader.memory[corrupt_addrs[arch][0] + i] = c
        boned = p.factory.simulation_manager().explore(find=target_addrs[arch], avoid=avoid_addrs[arch])
        assert len(boned.errored) >= 1
        assert isinstance(boned.errored[0].error, angr.SimIRSBNoDecodeError)
        assert boned.errored[0].state.addr == corrupt_addrs[arch][0]

        # hook the instructions with the emulated stuff
        p.hook(
            corrupt_addrs[arch][0],
            corrupt_addrs[arch][2],
            length=len(corrupt_addrs[arch][1]),
        )
        results = p.factory.simulation_manager().explore(find=target_addrs[arch], avoid=avoid_addrs[arch])
        stdin = results.found[0].posix.dumps(0)
        assert b"\x00\x00\x00\x00\x00\x00\x00\x00\x00SOSNEAKY\x00" == stdin

    def _run_merge(self, arch):
        p = angr.Project(os.path.join(test_location, arch, "fauxware"), auto_load_libs=False)
        pg = p.factory.simulation_manager()
        pg.explore()

        # release the unmergable data
        for s in pg.deadended:
            s.release_plugin("fs")
            if 3 in s.posix.fd:
                s.posix.close(3)

        pg.merge(stash="deadended", merge_key=lambda s: s.addr)

        path = pg.deadended[[b"Welcome" in s for s in pg.mp_deadended.posix.dumps(1).mp_items].index(True)]
        yes, no = path.history.merge_conditions
        inp = path.posix.stdin.content[2][0]  # content of second packet
        try:
            assert b"SOSNEAKY" in path.solver.eval(inp, cast_to=bytes, extra_constraints=(yes,))
            assert b"SOSNEAKY" not in path.solver.eval(inp, cast_to=bytes, extra_constraints=(no,))
        except AssertionError:
            yes, no = no, yes
            assert b"SOSNEAKY" in path.solver.eval(inp, cast_to=bytes, extra_constraints=(yes,))
            assert b"SOSNEAKY" not in path.solver.eval(inp, cast_to=bytes, extra_constraints=(no,))

    def test_merge_i386(self):
        self._run_merge("i386")

    def test_merge_x86_64(self):
        self._run_merge("x86_64")

    def test_merge_ppc(self):
        self._run_merge("ppc")

    def test_merge_armel(self):
        self._run_merge("armel")

    def test_merge_android(self):
        self._run_merge("android/arm")

    def test_merge_mips(self):
        self._run_merge("mips")

    def test_fauxware_i386(self):
        self._run_fauxware("i386")

    def test_fauxware_x86_64(self):
        self._run_fauxware("x86_64")

    def test_fauxware_ppc(self):
        self._run_fauxware("ppc")

    def test_fauxware_armel(self):
        self._run_fauxware("armel")

    def test_fauxware_android(self):
        self._run_fauxware("android/arm")

    def test_fauxware_mips(self):
        self._run_fauxware("mips")

    def test_pickling_i386(self):
        self._run_pickling("i386")

    def test_pickling_x86_64(self):
        self._run_pickling("x86_64")

    def test_pickling_ppc(self):
        self._run_pickling("ppc")

    def test_pickling_armel(self):
        self._run_pickling("armel")

    def test_pickling_mips(self):
        self._run_pickling("mips")

    @slow_test
    def test_fastmen(self):
        self._run_fastmem("x86_64")

    def test_nodecode_i386(self):
        self._run_nodecode("i386")

    def test_nodecode_x86_64(self):
        self._run_nodecode("x86_64")

    def test_nodecode_ppc(self):
        self._run_nodecode("ppc")

    def test_nodecode_armel(self):
        self._run_nodecode("armel")

    def test_nodecode_mips(self):
        self._run_nodecode("mips")


if __name__ == "__main__":
    unittest.main()



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("cryptxml.dll")
prototypes = \
    {
        #
        'CryptXmlClose': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hCryptXml"]),
        #
        'CryptXmlGetTransforms': SimTypeFunction([SimTypePointer(SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "cTransformInfo": SimTypeInt(signed=False, label="UInt32"), "rgpTransformInfo": SimTypePointer(SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithm": SimTypePointer(SimTypeChar(label="Char"), offset=0), "cbBufferSize": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="CRYPT_XML_TRANSFORM_FLAGS"), "pfnCreateTransform": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="CRYPT_XML_ALGORITHM"), offset=0), SimTypePointer(SimStruct({"pvCallbackState": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbBufferSize": SimTypeInt(signed=False, label="UInt32"), "pfnRead": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pvCallbackState", "pbData", "cbData", "pcbRead"]), offset=0), "pfnClose": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pvCallbackState"]), offset=0)}, name="CRYPT_XML_DATA_PROVIDER", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"pvCallbackState": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbBufferSize": SimTypeInt(signed=False, label="UInt32"), "pfnRead": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pvCallbackState", "pbData", "cbData", "pcbRead"]), offset=0), "pfnClose": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pvCallbackState"]), offset=0)}, name="CRYPT_XML_DATA_PROVIDER", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pTransform", "pProviderIn", "pProviderOut"]), offset=0)}, name="CRYPT_XML_TRANSFORM_INFO", pack=False, align=None), offset=0), offset=0)}, name="CRYPT_XML_TRANSFORM_CHAIN_CONFIG", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppConfig"]),
        #
        'CryptXmlOpenToEncode': SimTypeFunction([SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "cTransformInfo": SimTypeInt(signed=False, label="UInt32"), "rgpTransformInfo": SimTypePointer(SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithm": SimTypePointer(SimTypeChar(label="Char"), offset=0), "cbBufferSize": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="CRYPT_XML_TRANSFORM_FLAGS"), "pfnCreateTransform": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="CRYPT_XML_ALGORITHM"), offset=0), SimTypePointer(SimStruct({"pvCallbackState": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbBufferSize": SimTypeInt(signed=False, label="UInt32"), "pfnRead": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pvCallbackState", "pbData", "cbData", "pcbRead"]), offset=0), "pfnClose": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pvCallbackState"]), offset=0)}, name="CRYPT_XML_DATA_PROVIDER", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"pvCallbackState": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbBufferSize": SimTypeInt(signed=False, label="UInt32"), "pfnRead": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pvCallbackState", "pbData", "cbData", "pcbRead"]), offset=0), "pfnClose": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pvCallbackState"]), offset=0)}, name="CRYPT_XML_DATA_PROVIDER", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pTransform", "pProviderIn", "pProviderOut"]), offset=0)}, name="CRYPT_XML_TRANSFORM_INFO", pack=False, align=None), offset=0), offset=0)}, name="CRYPT_XML_TRANSFORM_CHAIN_CONFIG", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="CRYPT_XML_FLAGS"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimStruct({"dwPropId": SimTypeInt(signed=False, label="CRYPT_XML_PROPERTY_ID"), "pvValue": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbValue": SimTypeInt(signed=False, label="UInt32")}, name="CRYPT_XML_PROPERTY", pack=False, align=None), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pConfig", "dwFlags", "wszId", "rgProperty", "cProperty", "pEncoded", "phSignature"]),
        #
        'CryptXmlOpenToDecode': SimTypeFunction([SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "cTransformInfo": SimTypeInt(signed=False, label="UInt32"), "rgpTransformInfo": SimTypePointer(SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithm": SimTypePointer(SimTypeChar(label="Char"), offset=0), "cbBufferSize": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="CRYPT_XML_TRANSFORM_FLAGS"), "pfnCreateTransform": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="CRYPT_XML_ALGORITHM"), offset=0), SimTypePointer(SimStruct({"pvCallbackState": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbBufferSize": SimTypeInt(signed=False, label="UInt32"), "pfnRead": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pvCallbackState", "pbData", "cbData", "pcbRead"]), offset=0), "pfnClose": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pvCallbackState"]), offset=0)}, name="CRYPT_XML_DATA_PROVIDER", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"pvCallbackState": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbBufferSize": SimTypeInt(signed=False, label="UInt32"), "pfnRead": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pvCallbackState", "pbData", "cbData", "pcbRead"]), offset=0), "pfnClose": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pvCallbackState"]), offset=0)}, name="CRYPT_XML_DATA_PROVIDER", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pTransform", "pProviderIn", "pProviderOut"]), offset=0)}, name="CRYPT_XML_TRANSFORM_INFO", pack=False, align=None), offset=0), offset=0)}, name="CRYPT_XML_TRANSFORM_CHAIN_CONFIG", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="CRYPT_XML_FLAGS"), SimTypePointer(SimStruct({"dwPropId": SimTypeInt(signed=False, label="CRYPT_XML_PROPERTY_ID"), "pvValue": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbValue": SimTypeInt(signed=False, label="UInt32")}, name="CRYPT_XML_PROPERTY", pack=False, align=None), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pConfig", "dwFlags", "rgProperty", "cProperty", "pEncoded", "phCryptXml"]),
        #
        'CryptXmlAddObject': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"dwPropId": SimTypeInt(signed=False, label="CRYPT_XML_PROPERTY_ID"), "pvValue": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbValue": SimTypeInt(signed=False, label="UInt32")}, name="CRYPT_XML_PROPERTY", pack=False, align=None), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None), offset=0), SimTypePointer(SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "hObject": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "wszId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszMimeType": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszEncoding": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Manifest": SimStruct({"cReference": SimTypeInt(signed=False, label="UInt32"), "rgpReference": SimTypePointer(SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "hReference": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "wszId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszUri": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszType": SimTypePointer(SimTypeChar(label="Char"), offset=0), "DigestMethod": SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithm": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_ALGORITHM", pack=False, align=None), "DigestValue": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPTOAPI_BLOB", pack=False, align=None), "cTransform": SimTypeInt(signed=False, label="UInt32"), "rgTransform": SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithm": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_ALGORITHM", pack=False, align=None), offset=0)}, name="CRYPT_XML_REFERENCE", pack=False, align=None), offset=0), offset=0)}, name="CRYPT_XML_REFERENCES", pack=False, align=None), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_OBJECT", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hSignatureOrObject", "dwFlags", "rgProperty", "cProperty", "pEncoded", "ppObject"]),
        #
        'CryptXmlCreateReference': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithm": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_ALGORITHM", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithm": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_ALGORITHM", pack=False, align=None), label="LPArray", offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hCryptXml", "dwFlags", "wszId", "wszURI", "wszType", "pDigestMethod", "cTransform", "rgTransform", "phReference"]),
        #
        'CryptXmlDigestReference': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"pvCallbackState": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbBufferSize": SimTypeInt(signed=False, label="UInt32"), "pfnRead": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pvCallbackState", "pbData", "cbData", "pcbRead"]), offset=0), "pfnClose": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pvCallbackState"]), offset=0)}, name="CRYPT_XML_DATA_PROVIDER", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hReference", "dwFlags", "pDataProviderIn"]),
        #
        'CryptXmlSetHMACSecret': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hSignature", "pbSecret", "cbSecret"]),
        #
        'CryptXmlSign': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeInt(signed=False, label="CERT_KEY_SPEC"), SimTypeInt(signed=False, label="CRYPT_XML_FLAGS"), SimTypeInt(signed=False, label="CRYPT_XML_KEYINFO_SPEC"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithm": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_ALGORITHM", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithm": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_ALGORITHM", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hSignature", "hKey", "dwKeySpec", "dwFlags", "dwKeyInfoSpec", "pvKeyInfoSpec", "pSignatureMethod", "pCanonicalization"]),
        #
        'CryptXmlImportPublicKey': SimTypeFunction([SimTypeInt(signed=False, label="CRYPT_XML_FLAGS"), SimTypePointer(SimStruct({"dwType": SimTypeInt(signed=False, label="CRYPT_XML_KEY_VALUE_TYPE"), "Anonymous": SimUnion({"DSAKeyValue": SimStruct({"P": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "Q": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "G": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "Y": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "J": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "Seed": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "Counter": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None)}, name="CRYPT_XML_KEY_DSA_KEY_VALUE", pack=False, align=None), "RSAKeyValue": SimStruct({"Modulus": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "Exponent": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None)}, name="CRYPT_XML_KEY_RSA_KEY_VALUE", pack=False, align=None), "ECDSAKeyValue": SimStruct({"wszNamedCurve": SimTypePointer(SimTypeChar(label="Char"), offset=0), "X": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "Y": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "ExplicitPara": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_KEY_ECDSA_KEY_VALUE", pack=False, align=None), "Custom": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="<anon>", label="None")}, name="CRYPT_XML_KEY_VALUE", pack=False, align=None), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["dwFlags", "pKeyValue", "phKey"]),
        #
        'CryptXmlVerifySignature': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="CRYPT_XML_FLAGS")], SimTypeInt(signed=True, label="Int32"), arg_names=["hSignature", "hKey", "dwFlags"]),
        #
        'CryptXmlGetDocContext': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "hDocCtxt": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "pTransformsConfig": SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "cTransformInfo": SimTypeInt(signed=False, label="UInt32"), "rgpTransformInfo": SimTypePointer(SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithm": SimTypePointer(SimTypeChar(label="Char"), offset=0), "cbBufferSize": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="CRYPT_XML_TRANSFORM_FLAGS"), "pfnCreateTransform": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="CRYPT_XML_ALGORITHM"), offset=0), SimTypePointer(SimStruct({"pvCallbackState": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbBufferSize": SimTypeInt(signed=False, label="UInt32"), "pfnRead": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pvCallbackState", "pbData", "cbData", "pcbRead"]), offset=0), "pfnClose": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pvCallbackState"]), offset=0)}, name="CRYPT_XML_DATA_PROVIDER", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"pvCallbackState": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbBufferSize": SimTypeInt(signed=False, label="UInt32"), "pfnRead": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pvCallbackState", "pbData", "cbData", "pcbRead"]), offset=0), "pfnClose": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pvCallbackState"]), offset=0)}, name="CRYPT_XML_DATA_PROVIDER", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pTransform", "pProviderIn", "pProviderOut"]), offset=0)}, name="CRYPT_XML_TRANSFORM_INFO", pack=False, align=None), offset=0), offset=0)}, name="CRYPT_XML_TRANSFORM_CHAIN_CONFIG", pack=False, align=None), offset=0), "cSignature": SimTypeInt(signed=False, label="UInt32"), "rgpSignature": SimTypePointer(SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "hSignature": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "wszId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "SignedInfo": SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Canonicalization": SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithm": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_ALGORITHM", pack=False, align=None), "SignatureMethod": SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithm": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_ALGORITHM", pack=False, align=None), "cReference": SimTypeInt(signed=False, label="UInt32"), "rgpReference": SimTypePointer(SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "hReference": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "wszId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszUri": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszType": SimTypePointer(SimTypeChar(label="Char"), offset=0), "DigestMethod": SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithm": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_ALGORITHM", pack=False, align=None), "DigestValue": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPTOAPI_BLOB", pack=False, align=None), "cTransform": SimTypeInt(signed=False, label="UInt32"), "rgTransform": SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithm": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_ALGORITHM", pack=False, align=None), offset=0)}, name="CRYPT_XML_REFERENCE", pack=False, align=None), offset=0), offset=0), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_SIGNED_INFO", pack=False, align=None), "SignatureValue": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPTOAPI_BLOB", pack=False, align=None), "pKeyInfo": SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "cKeyInfo": SimTypeInt(signed=False, label="UInt32"), "rgKeyInfo": SimTypePointer(SimStruct({"dwType": SimTypeInt(signed=False, label="CRYPT_XML_KEYINFO_TYPE"), "Anonymous": SimUnion({"wszKeyName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "KeyValue": SimStruct({"dwType": SimTypeInt(signed=False, label="CRYPT_XML_KEY_VALUE_TYPE"), "Anonymous": SimUnion({"DSAKeyValue": SimStruct({"P": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "Q": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "G": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "Y": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "J": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "Seed": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "Counter": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None)}, name="CRYPT_XML_KEY_DSA_KEY_VALUE", pack=False, align=None), "RSAKeyValue": SimStruct({"Modulus": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "Exponent": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None)}, name="CRYPT_XML_KEY_RSA_KEY_VALUE", pack=False, align=None), "ECDSAKeyValue": SimStruct({"wszNamedCurve": SimTypePointer(SimTypeChar(label="Char"), offset=0), "X": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "Y": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "ExplicitPara": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_KEY_ECDSA_KEY_VALUE", pack=False, align=None), "Custom": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="<anon>", label="None")}, name="CRYPT_XML_KEY_VALUE", pack=False, align=None), "RetrievalMethod": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None), "X509Data": SimStruct({"cX509Data": SimTypeInt(signed=False, label="UInt32"), "rgX509Data": SimTypePointer(SimStruct({"dwType": SimTypeInt(signed=False, label="CRYPT_XML_X509DATA_TYPE"), "Anonymous": SimUnion({"IssuerSerial": SimStruct({"wszIssuer": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszSerial": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="CRYPT_XML_ISSUER_SERIAL", pack=False, align=None), "SKI": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "wszSubjectName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Certificate": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "CRL": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "Custom": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="<anon>", label="None")}, name="CRYPT_XML_X509DATA_ITEM", pack=False, align=None), offset=0)}, name="CRYPT_XML_X509DATA", pack=False, align=None), "Custom": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="<anon>", label="None")}, name="CRYPT_XML_KEY_INFO_ITEM", pack=False, align=None), offset=0), "hVerifyKey": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="CRYPT_XML_KEY_INFO", pack=False, align=None), offset=0), "cObject": SimTypeInt(signed=False, label="UInt32"), "rgpObject": SimTypePointer(SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "hObject": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "wszId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszMimeType": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszEncoding": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Manifest": SimStruct({"cReference": SimTypeInt(signed=False, label="UInt32"), "rgpReference": SimTypePointer(SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "hReference": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "wszId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszUri": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszType": SimTypePointer(SimTypeChar(label="Char"), offset=0), "DigestMethod": SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithm": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_ALGORITHM", pack=False, align=None), "DigestValue": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPTOAPI_BLOB", pack=False, align=None), "cTransform": SimTypeInt(signed=False, label="UInt32"), "rgTransform": SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithm": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_ALGORITHM", pack=False, align=None), offset=0)}, name="CRYPT_XML_REFERENCE", pack=False, align=None), offset=0), offset=0)}, name="CRYPT_XML_REFERENCES", pack=False, align=None), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_OBJECT", pack=False, align=None), offset=0), offset=0)}, name="CRYPT_XML_SIGNATURE", pack=False, align=None), offset=0), offset=0)}, name="CRYPT_XML_DOC_CTXT", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hCryptXml", "ppStruct"]),
        #
        'CryptXmlGetSignature': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "hSignature": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "wszId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "SignedInfo": SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Canonicalization": SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithm": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_ALGORITHM", pack=False, align=None), "SignatureMethod": SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithm": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_ALGORITHM", pack=False, align=None), "cReference": SimTypeInt(signed=False, label="UInt32"), "rgpReference": SimTypePointer(SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "hReference": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "wszId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszUri": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszType": SimTypePointer(SimTypeChar(label="Char"), offset=0), "DigestMethod": SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithm": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_ALGORITHM", pack=False, align=None), "DigestValue": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPTOAPI_BLOB", pack=False, align=None), "cTransform": SimTypeInt(signed=False, label="UInt32"), "rgTransform": SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithm": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_ALGORITHM", pack=False, align=None), offset=0)}, name="CRYPT_XML_REFERENCE", pack=False, align=None), offset=0), offset=0), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_SIGNED_INFO", pack=False, align=None), "SignatureValue": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPTOAPI_BLOB", pack=False, align=None), "pKeyInfo": SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "cKeyInfo": SimTypeInt(signed=False, label="UInt32"), "rgKeyInfo": SimTypePointer(SimStruct({"dwType": SimTypeInt(signed=False, label="CRYPT_XML_KEYINFO_TYPE"), "Anonymous": SimUnion({"wszKeyName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "KeyValue": SimStruct({"dwType": SimTypeInt(signed=False, label="CRYPT_XML_KEY_VALUE_TYPE"), "Anonymous": SimUnion({"DSAKeyValue": SimStruct({"P": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "Q": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "G": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "Y": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "J": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "Seed": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "Counter": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None)}, name="CRYPT_XML_KEY_DSA_KEY_VALUE", pack=False, align=None), "RSAKeyValue": SimStruct({"Modulus": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "Exponent": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None)}, name="CRYPT_XML_KEY_RSA_KEY_VALUE", pack=False, align=None), "ECDSAKeyValue": SimStruct({"wszNamedCurve": SimTypePointer(SimTypeChar(label="Char"), offset=0), "X": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "Y": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "ExplicitPara": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_KEY_ECDSA_KEY_VALUE", pack=False, align=None), "Custom": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="<anon>", label="None")}, name="CRYPT_XML_KEY_VALUE", pack=False, align=None), "RetrievalMethod": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None), "X509Data": SimStruct({"cX509Data": SimTypeInt(signed=False, label="UInt32"), "rgX509Data": SimTypePointer(SimStruct({"dwType": SimTypeInt(signed=False, label="CRYPT_XML_X509DATA_TYPE"), "Anonymous": SimUnion({"IssuerSerial": SimStruct({"wszIssuer": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszSerial": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="CRYPT_XML_ISSUER_SERIAL", pack=False, align=None), "SKI": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "wszSubjectName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Certificate": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "CRL": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_DATA_BLOB", pack=False, align=None), "Custom": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="<anon>", label="None")}, name="CRYPT_XML_X509DATA_ITEM", pack=False, align=None), offset=0)}, name="CRYPT_XML_X509DATA", pack=False, align=None), "Custom": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="<anon>", label="None")}, name="CRYPT_XML_KEY_INFO_ITEM", pack=False, align=None), offset=0), "hVerifyKey": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="CRYPT_XML_KEY_INFO", pack=False, align=None), offset=0), "cObject": SimTypeInt(signed=False, label="UInt32"), "rgpObject": SimTypePointer(SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "hObject": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "wszId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszMimeType": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszEncoding": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Manifest": SimStruct({"cReference": SimTypeInt(signed=False, label="UInt32"), "rgpReference": SimTypePointer(SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "hReference": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "wszId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszUri": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszType": SimTypePointer(SimTypeChar(label="Char"), offset=0), "DigestMethod": SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithm": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_ALGORITHM", pack=False, align=None), "DigestValue": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPTOAPI_BLOB", pack=False, align=None), "cTransform": SimTypeInt(signed=False, label="UInt32"), "rgTransform": SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithm": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_ALGORITHM", pack=False, align=None), offset=0)}, name="CRYPT_XML_REFERENCE", pack=False, align=None), offset=0), offset=0)}, name="CRYPT_XML_REFERENCES", pack=False, align=None), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_OBJECT", pack=False, align=None), offset=0), offset=0)}, name="CRYPT_XML_SIGNATURE", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hCryptXml", "ppStruct"]),
        #
        'CryptXmlGetReference': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "hReference": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "wszId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszUri": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszType": SimTypePointer(SimTypeChar(label="Char"), offset=0), "DigestMethod": SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithm": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_ALGORITHM", pack=False, align=None), "DigestValue": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPTOAPI_BLOB", pack=False, align=None), "cTransform": SimTypeInt(signed=False, label="UInt32"), "rgTransform": SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithm": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_ALGORITHM", pack=False, align=None), offset=0)}, name="CRYPT_XML_REFERENCE", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hCryptXml", "ppStruct"]),
        #
        'CryptXmlGetStatus': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "dwErrorStatus": SimTypeInt(signed=False, label="CRYPT_XML_STATUS_ERROR_STATUS"), "dwInfoStatus": SimTypeInt(signed=False, label="CRYPT_XML_STATUS_INFO_STATUS")}, name="CRYPT_XML_STATUS", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hCryptXml", "pStatus"]),
        #
        'CryptXmlEncode': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), SimTypePointer(SimStruct({"dwPropId": SimTypeInt(signed=False, label="CRYPT_XML_PROPERTY_ID"), "pvValue": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbValue": SimTypeInt(signed=False, label="UInt32")}, name="CRYPT_XML_PROPERTY", pack=False, align=None), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pvCallbackState", "pbData", "cbData"]), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hCryptXml", "dwCharset", "rgProperty", "cProperty", "pvCallbackState", "pfnWrite"]),
        #
        'CryptXmlGetAlgorithmInfo': SimTypeFunction([SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithm": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Encoded": SimStruct({"dwCharset": SimTypeInt(signed=False, label="CRYPT_XML_CHARSET"), "cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPT_XML_BLOB", pack=False, align=None)}, name="CRYPT_XML_ALGORITHM", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="CRYPT_XML_FLAGS"), SimTypePointer(SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithmURI": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwGroupId": SimTypeInt(signed=False, label="CRYPT_XML_GROUP_ID"), "wszCNGAlgid": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszCNGExtraAlgid": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwSignFlags": SimTypeInt(signed=False, label="UInt32"), "dwVerifyFlags": SimTypeInt(signed=False, label="UInt32"), "pvPaddingInfo": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "pvExtraInfo": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="CRYPT_XML_ALGORITHM_INFO", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pXmlAlgorithm", "dwFlags", "ppAlgInfo"]),
        #
        'CryptXmlFindAlgorithmInfo': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithmURI": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwGroupId": SimTypeInt(signed=False, label="CRYPT_XML_GROUP_ID"), "wszCNGAlgid": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszCNGExtraAlgid": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwSignFlags": SimTypeInt(signed=False, label="UInt32"), "dwVerifyFlags": SimTypeInt(signed=False, label="UInt32"), "pvPaddingInfo": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "pvExtraInfo": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="CRYPT_XML_ALGORITHM_INFO", pack=False, align=None), offset=0), arg_names=["dwFindByType", "pvFindBy", "dwGroupId", "dwFlags"]),
        #
        'CryptXmlEnumAlgorithmInfo': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeFunction([SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wszAlgorithmURI": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwGroupId": SimTypeInt(signed=False, label="CRYPT_XML_GROUP_ID"), "wszCNGAlgid": SimTypePointer(SimTypeChar(label="Char"), offset=0), "wszCNGExtraAlgid": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwSignFlags": SimTypeInt(signed=False, label="UInt32"), "dwVerifyFlags": SimTypeInt(signed=False, label="UInt32"), "pvPaddingInfo": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "pvExtraInfo": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="CRYPT_XML_ALGORITHM_INFO", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pInfo", "pvArg"]), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["dwGroupId", "dwFlags", "pvArg", "pfnEnumAlgInfo"]),
    }

lib.set_prototypes(prototypes)



#!/usr/bin/env python3
# pylint: disable=missing-class-docstring,no-self-use,line-too-long

import unittest

import claripy

import angr

# all the input values were generated via
# [random.randrange(256) for _ in range(16)]
# then set into the input registers via gdb
# set $xmm0.v16_int8 = {...}
# then read out as uint128s
# p/x $xmm0.uint128
# then single stepped and the result read out


class TestOps(unittest.TestCase):
    def test_irop_perm(self):
        p = angr.load_shellcode("vpshufb xmm0,xmm1,xmm2", "amd64")

        # concrete test
        s1 = p.factory.blank_state()
        s1.regs.xmm1 = 0x3C899A56814EE9B84C7B5D8394C85881
        s1.regs.xmm2 = 0xA55C66A2CDEF1CBCD72B42078D1B7F8B
        s2 = s1.step(num_inst=1).successors[0]
        assert (s2.regs.xmm0 == 0x00567B00000056000081C84C00813C00).is_true()

        # symbolic test
        s3 = p.factory.blank_state()
        s3.regs.xmm1 = claripy.BVS("xmm1", 128)
        s3.regs.xmm2 = claripy.BVS("xmm2", 128)
        s4 = s3.step(num_inst=1).successors[0]
        s4.solver.add(s4.regs.xmm2 == 0xA55C66A2CDEF1CBCD72B42078D1B7F8B)
        s4.solver.add(s4.regs.xmm0 == 0x00567B00000056000081C84C00813C00)
        assert s4.solver.solution(s4.regs.xmm1, 0x3C899A56814EE9B84C7B5D8394C85881)

    def test_irop_mulhi(self):
        p = angr.load_shellcode("vpmulhw xmm0,xmm1,xmm2", "amd64")

        # concrete test
        s1 = p.factory.blank_state()
        s1.regs.xmm1 = 0x3ACA92553C2526D4F20987AEAB250255
        s1.regs.xmm2 = 0x1AEBCB281463274EC3CE6473619A8541
        s2 = s1.step(num_inst=1).successors[0]
        assert (s2.regs.xmm0 == 0x62E16A304CA05F60348D0C9DFA5FEE1).is_true()

    def test_irop_catevenlanes(self):
        p = angr.load_shellcode("pmulhrsw xmm0, xmm1", "amd64")

        # concrete test
        s1 = p.factory.blank_state()
        s1.regs.xmm0 = 0x4713E06BF3235E97CA8CFDE0647D65FD
        s1.regs.xmm1 = 0x31F1F86DA1DCE7DE252ADC78160E1016
        s2 = s1.step(num_inst=1).successors[0]
        assert (s2.regs.xmm0 == 0x1BBB01DE0976EE2BF07B009711500CD1).is_true()

    def test_saturating_packing(self):
        # SaturateSignedWordToUnsignedByte
        p = angr.load_shellcode("vpackuswb xmm1, xmm0, xmm0", arch="amd64")
        s = p.factory.blank_state()
        s.regs.xmm0 = 0x0000_0001_7FFE_7FFF_8000_8001_FFFE_FFFF
        s = s.step(num_inst=1).successors[0]
        assert (s.regs.xmm1 == 0x00_01_FF_FF_00_00_00_00_0001FFFF00000000).is_true()

        # "Pack with unsigned saturation"
        p = angr.load_shellcode("vpackusdw xmm1, xmm0, xmm0", arch="amd64")
        s = p.factory.blank_state()
        s.regs.xmm0 = 0x00000001_7FFFFFFE_80000001_FFFFFFFE
        s = s.step(num_inst=1).successors[0]
        assert (s.regs.xmm1 == 0x0001_FFFF_0000_0000_0001FFFF00000000).is_true()

        # SaturateSignedWordToSignedByte
        p = angr.load_shellcode("vpacksswb xmm1, xmm0, xmm0", arch="amd64")
        s = p.factory.blank_state()
        s.regs.xmm0 = 0x0000_0001_7FFE_7FFF_8000_8001_FFFE_FFFF
        s = s.step(num_inst=1).successors[0]
        assert (s.regs.xmm1 == 0x00_01_7F_7F_80_80_FE_FF_00017F7F8080FEFF).is_true()


if __name__ == "__main__":
    unittest.main()



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("winhvplatform.dll")
prototypes = \
    {
        #
        'WHvGetCapability': SimTypeFunction([SimTypeInt(signed=False, label="WHV_CAPABILITY_CODE"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["CapabilityCode", "CapabilityBuffer", "CapabilityBufferSizeInBytes", "WrittenSizeInBytes"]),
        #
        'WHvCreatePartition': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition"]),
        #
        'WHvSetupPartition': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition"]),
        #
        'WHvDeletePartition': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition"]),
        #
        'WHvGetPartitionProperty': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="WHV_PARTITION_PROPERTY_CODE"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition", "PropertyCode", "PropertyBuffer", "PropertyBufferSizeInBytes", "WrittenSizeInBytes"]),
        #
        'WHvSetPartitionProperty': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="WHV_PARTITION_PROPERTY_CODE"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition", "PropertyCode", "PropertyBuffer", "PropertyBufferSizeInBytes"]),
        #
        'WHvSuspendPartitionTime': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition"]),
        #
        'WHvResumePartitionTime': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition"]),
        #
        'WHvMapGpaRange': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeLongLong(signed=False, label="UInt64"), SimTypeLongLong(signed=False, label="UInt64"), SimTypeInt(signed=False, label="WHV_MAP_GPA_RANGE_FLAGS")], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition", "SourceAddress", "GuestAddress", "SizeInBytes", "Flags"]),
        #
        'WHvUnmapGpaRange': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeLongLong(signed=False, label="UInt64"), SimTypeLongLong(signed=False, label="UInt64")], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition", "GuestAddress", "SizeInBytes"]),
        #
        'WHvTranslateGva': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeLongLong(signed=False, label="UInt64"), SimTypeInt(signed=False, label="WHV_TRANSLATE_GVA_FLAGS"), SimTypePointer(SimStruct({"ResultCode": SimTypeInt(signed=False, label="WHV_TRANSLATE_GVA_RESULT_CODE"), "Reserved": SimTypeInt(signed=False, label="UInt32")}, name="WHV_TRANSLATE_GVA_RESULT", pack=False, align=None), offset=0), SimTypePointer(SimTypeLongLong(signed=False, label="UInt64"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition", "VpIndex", "Gva", "TranslateFlags", "TranslationResult", "Gpa"]),
        #
        'WHvCreateVirtualProcessor': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition", "VpIndex", "Flags"]),
        #
        'WHvDeleteVirtualProcessor': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition", "VpIndex"]),
        #
        'WHvRunVirtualProcessor': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition", "VpIndex", "ExitContext", "ExitContextSizeInBytes"]),
        #
        'WHvCancelRunVirtualProcessor': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition", "VpIndex", "Flags"]),
        #
        'WHvGetVirtualProcessorRegisters': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="WHV_REGISTER_NAME"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimUnion({"Reg128": SimUnion({"Anonymous": SimStruct({"Low64": SimTypeLongLong(signed=False, label="UInt64"), "High64": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None), "Dword": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4)}, name="<anon>", label="None"), "Reg64": SimTypeLongLong(signed=False, label="UInt64"), "Reg32": SimTypeInt(signed=False, label="UInt32"), "Reg16": SimTypeShort(signed=False, label="UInt16"), "Reg8": SimTypeChar(label="Byte"), "Fp": SimUnion({"Anonymous": SimStruct({"Mantissa": SimTypeLongLong(signed=False, label="UInt64"), "_bitfield": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None), "AsUINT128": SimUnion({"Anonymous": SimStruct({"Low64": SimTypeLongLong(signed=False, label="UInt64"), "High64": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None), "Dword": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4)}, name="<anon>", label="None")}, name="<anon>", label="None"), "FpControlStatus": SimUnion({"Anonymous": SimStruct({"FpControl": SimTypeShort(signed=False, label="UInt16"), "FpStatus": SimTypeShort(signed=False, label="UInt16"), "FpTag": SimTypeChar(label="Byte"), "Reserved": SimTypeChar(label="Byte"), "LastFpOp": SimTypeShort(signed=False, label="UInt16"), "Anonymous": SimUnion({"LastFpRip": SimTypeLongLong(signed=False, label="UInt64"), "Anonymous": SimStruct({"LastFpEip": SimTypeInt(signed=False, label="UInt32"), "LastFpCs": SimTypeShort(signed=False, label="UInt16"), "Reserved2": SimTypeShort(signed=False, label="UInt16")}, name="_Anonymous_e__Struct", pack=False, align=None)}, name="<anon>", label="None")}, name="_Anonymous_e__Struct", pack=False, align=None), "AsUINT128": SimUnion({"Anonymous": SimStruct({"Low64": SimTypeLongLong(signed=False, label="UInt64"), "High64": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None), "Dword": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4)}, name="<anon>", label="None")}, name="<anon>", label="None"), "XmmControlStatus": SimUnion({"Anonymous": SimStruct({"Anonymous": SimUnion({"LastFpRdp": SimTypeLongLong(signed=False, label="UInt64"), "Anonymous": SimStruct({"LastFpDp": SimTypeInt(signed=False, label="UInt32"), "LastFpDs": SimTypeShort(signed=False, label="UInt16"), "Reserved": SimTypeShort(signed=False, label="UInt16")}, name="_Anonymous_e__Struct", pack=False, align=None)}, name="<anon>", label="None"), "XmmStatusControl": SimTypeInt(signed=False, label="UInt32"), "XmmStatusControlMask": SimTypeInt(signed=False, label="UInt32")}, name="_Anonymous_e__Struct", pack=False, align=None), "AsUINT128": SimUnion({"Anonymous": SimStruct({"Low64": SimTypeLongLong(signed=False, label="UInt64"), "High64": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None), "Dword": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4)}, name="<anon>", label="None")}, name="<anon>", label="None"), "Segment": SimStruct({"Base": SimTypeLongLong(signed=False, label="UInt64"), "Limit": SimTypeInt(signed=False, label="UInt32"), "Selector": SimTypeShort(signed=False, label="UInt16"), "Anonymous": SimUnion({"Anonymous": SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="_Anonymous_e__Struct", pack=False, align=None), "Attributes": SimTypeShort(signed=False, label="UInt16")}, name="<anon>", label="None")}, name="WHV_X64_SEGMENT_REGISTER", pack=False, align=None), "Table": SimStruct({"Pad": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 3), "Limit": SimTypeShort(signed=False, label="UInt16"), "Base": SimTypeLongLong(signed=False, label="UInt64")}, name="WHV_X64_TABLE_REGISTER", pack=False, align=None), "InterruptState": SimUnion({"Anonymous": SimStruct({"_bitfield": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None), "AsUINT64": SimTypeLongLong(signed=False, label="UInt64")}, name="<anon>", label="None"), "PendingInterruption": SimUnion({"Anonymous": SimStruct({"_bitfield": SimTypeInt(signed=False, label="UInt32"), "ErrorCode": SimTypeInt(signed=False, label="UInt32")}, name="_Anonymous_e__Struct", pack=False, align=None), "AsUINT64": SimTypeLongLong(signed=False, label="UInt64")}, name="<anon>", label="None"), "DeliverabilityNotifications": SimUnion({"Anonymous": SimStruct({"_bitfield": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None), "AsUINT64": SimTypeLongLong(signed=False, label="UInt64")}, name="<anon>", label="None"), "ExceptionEvent": SimUnion({"Anonymous": SimStruct({"_bitfield": SimTypeInt(signed=False, label="UInt32"), "ErrorCode": SimTypeInt(signed=False, label="UInt32"), "ExceptionParameter": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None), "AsUINT128": SimUnion({"Anonymous": SimStruct({"Low64": SimTypeLongLong(signed=False, label="UInt64"), "High64": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None), "Dword": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4)}, name="<anon>", label="None")}, name="<anon>", label="None"), "ExtIntEvent": SimUnion({"Anonymous": SimStruct({"_bitfield": SimTypeLongLong(signed=False, label="UInt64"), "Reserved2": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None), "AsUINT128": SimUnion({"Anonymous": SimStruct({"Low64": SimTypeLongLong(signed=False, label="UInt64"), "High64": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None), "Dword": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4)}, name="<anon>", label="None")}, name="<anon>", label="None"), "InternalActivity": SimUnion({"Anonymous": SimStruct({"_bitfield": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None), "AsUINT64": SimTypeLongLong(signed=False, label="UInt64")}, name="<anon>", label="None"), "PendingDebugException": SimUnion({"AsUINT64": SimTypeLongLong(signed=False, label="UInt64"), "Anonymous": SimStruct({"_bitfield": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None)}, name="<anon>", label="None")}, name="<anon>", label="None"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition", "VpIndex", "RegisterNames", "RegisterCount", "RegisterValues"]),
        #
        'WHvSetVirtualProcessorRegisters': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="WHV_REGISTER_NAME"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimUnion({"Reg128": SimUnion({"Anonymous": SimStruct({"Low64": SimTypeLongLong(signed=False, label="UInt64"), "High64": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None), "Dword": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4)}, name="<anon>", label="None"), "Reg64": SimTypeLongLong(signed=False, label="UInt64"), "Reg32": SimTypeInt(signed=False, label="UInt32"), "Reg16": SimTypeShort(signed=False, label="UInt16"), "Reg8": SimTypeChar(label="Byte"), "Fp": SimUnion({"Anonymous": SimStruct({"Mantissa": SimTypeLongLong(signed=False, label="UInt64"), "_bitfield": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None), "AsUINT128": SimUnion({"Anonymous": SimStruct({"Low64": SimTypeLongLong(signed=False, label="UInt64"), "High64": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None), "Dword": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4)}, name="<anon>", label="None")}, name="<anon>", label="None"), "FpControlStatus": SimUnion({"Anonymous": SimStruct({"FpControl": SimTypeShort(signed=False, label="UInt16"), "FpStatus": SimTypeShort(signed=False, label="UInt16"), "FpTag": SimTypeChar(label="Byte"), "Reserved": SimTypeChar(label="Byte"), "LastFpOp": SimTypeShort(signed=False, label="UInt16"), "Anonymous": SimUnion({"LastFpRip": SimTypeLongLong(signed=False, label="UInt64"), "Anonymous": SimStruct({"LastFpEip": SimTypeInt(signed=False, label="UInt32"), "LastFpCs": SimTypeShort(signed=False, label="UInt16"), "Reserved2": SimTypeShort(signed=False, label="UInt16")}, name="_Anonymous_e__Struct", pack=False, align=None)}, name="<anon>", label="None")}, name="_Anonymous_e__Struct", pack=False, align=None), "AsUINT128": SimUnion({"Anonymous": SimStruct({"Low64": SimTypeLongLong(signed=False, label="UInt64"), "High64": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None), "Dword": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4)}, name="<anon>", label="None")}, name="<anon>", label="None"), "XmmControlStatus": SimUnion({"Anonymous": SimStruct({"Anonymous": SimUnion({"LastFpRdp": SimTypeLongLong(signed=False, label="UInt64"), "Anonymous": SimStruct({"LastFpDp": SimTypeInt(signed=False, label="UInt32"), "LastFpDs": SimTypeShort(signed=False, label="UInt16"), "Reserved": SimTypeShort(signed=False, label="UInt16")}, name="_Anonymous_e__Struct", pack=False, align=None)}, name="<anon>", label="None"), "XmmStatusControl": SimTypeInt(signed=False, label="UInt32"), "XmmStatusControlMask": SimTypeInt(signed=False, label="UInt32")}, name="_Anonymous_e__Struct", pack=False, align=None), "AsUINT128": SimUnion({"Anonymous": SimStruct({"Low64": SimTypeLongLong(signed=False, label="UInt64"), "High64": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None), "Dword": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4)}, name="<anon>", label="None")}, name="<anon>", label="None"), "Segment": SimStruct({"Base": SimTypeLongLong(signed=False, label="UInt64"), "Limit": SimTypeInt(signed=False, label="UInt32"), "Selector": SimTypeShort(signed=False, label="UInt16"), "Anonymous": SimUnion({"Anonymous": SimStruct({"_bitfield": SimTypeShort(signed=False, label="UInt16")}, name="_Anonymous_e__Struct", pack=False, align=None), "Attributes": SimTypeShort(signed=False, label="UInt16")}, name="<anon>", label="None")}, name="WHV_X64_SEGMENT_REGISTER", pack=False, align=None), "Table": SimStruct({"Pad": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 3), "Limit": SimTypeShort(signed=False, label="UInt16"), "Base": SimTypeLongLong(signed=False, label="UInt64")}, name="WHV_X64_TABLE_REGISTER", pack=False, align=None), "InterruptState": SimUnion({"Anonymous": SimStruct({"_bitfield": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None), "AsUINT64": SimTypeLongLong(signed=False, label="UInt64")}, name="<anon>", label="None"), "PendingInterruption": SimUnion({"Anonymous": SimStruct({"_bitfield": SimTypeInt(signed=False, label="UInt32"), "ErrorCode": SimTypeInt(signed=False, label="UInt32")}, name="_Anonymous_e__Struct", pack=False, align=None), "AsUINT64": SimTypeLongLong(signed=False, label="UInt64")}, name="<anon>", label="None"), "DeliverabilityNotifications": SimUnion({"Anonymous": SimStruct({"_bitfield": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None), "AsUINT64": SimTypeLongLong(signed=False, label="UInt64")}, name="<anon>", label="None"), "ExceptionEvent": SimUnion({"Anonymous": SimStruct({"_bitfield": SimTypeInt(signed=False, label="UInt32"), "ErrorCode": SimTypeInt(signed=False, label="UInt32"), "ExceptionParameter": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None), "AsUINT128": SimUnion({"Anonymous": SimStruct({"Low64": SimTypeLongLong(signed=False, label="UInt64"), "High64": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None), "Dword": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4)}, name="<anon>", label="None")}, name="<anon>", label="None"), "ExtIntEvent": SimUnion({"Anonymous": SimStruct({"_bitfield": SimTypeLongLong(signed=False, label="UInt64"), "Reserved2": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None), "AsUINT128": SimUnion({"Anonymous": SimStruct({"Low64": SimTypeLongLong(signed=False, label="UInt64"), "High64": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None), "Dword": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4)}, name="<anon>", label="None")}, name="<anon>", label="None"), "InternalActivity": SimUnion({"Anonymous": SimStruct({"_bitfield": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None), "AsUINT64": SimTypeLongLong(signed=False, label="UInt64")}, name="<anon>", label="None"), "PendingDebugException": SimUnion({"AsUINT64": SimTypeLongLong(signed=False, label="UInt64"), "Anonymous": SimStruct({"_bitfield": SimTypeLongLong(signed=False, label="UInt64")}, name="_Anonymous_e__Struct", pack=False, align=None)}, name="<anon>", label="None")}, name="<anon>", label="None"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition", "VpIndex", "RegisterNames", "RegisterCount", "RegisterValues"]),
        #
        'WHvGetVirtualProcessorInterruptControllerState': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition", "VpIndex", "State", "StateSize", "WrittenSize"]),
        #
        'WHvSetVirtualProcessorInterruptControllerState': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition", "VpIndex", "State", "StateSize"]),
        #
        'WHvRequestInterrupt': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"_bitfield": SimTypeLongLong(signed=False, label="UInt64"), "Destination": SimTypeInt(signed=False, label="UInt32"), "Vector": SimTypeInt(signed=False, label="UInt32")}, name="WHV_INTERRUPT_CONTROL", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition", "Interrupt", "InterruptControlSize"]),
        #
        'WHvGetVirtualProcessorXsaveState': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition", "VpIndex", "Buffer", "BufferSizeInBytes", "BytesWritten"]),
        #
        'WHvSetVirtualProcessorXsaveState': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition", "VpIndex", "Buffer", "BufferSizeInBytes"]),
        #
        'WHvQueryGpaRangeDirtyBitmap': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeLongLong(signed=False, label="UInt64"), SimTypeLongLong(signed=False, label="UInt64"), SimTypePointer(SimTypeLongLong(signed=False, label="UInt64"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition", "GuestAddress", "RangeSizeInBytes", "Bitmap", "BitmapSizeInBytes"]),
        #
        'WHvGetPartitionCounters': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="WHV_PARTITION_COUNTER_SET"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition", "CounterSet", "Buffer", "BufferSizeInBytes", "BytesWritten"]),
        #
        'WHvGetVirtualProcessorCounters': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="WHV_PROCESSOR_COUNTER_SET"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition", "VpIndex", "CounterSet", "Buffer", "BufferSizeInBytes", "BytesWritten"]),
        #
        'WHvGetVirtualProcessorInterruptControllerState2': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition", "VpIndex", "State", "StateSize", "WrittenSize"]),
        #
        'WHvSetVirtualProcessorInterruptControllerState2': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition", "VpIndex", "State", "StateSize"]),
        #
        'WHvRegisterPartitionDoorbellEvent': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"GuestAddress": SimTypeLongLong(signed=False, label="UInt64"), "Value": SimTypeLongLong(signed=False, label="UInt64"), "Length": SimTypeInt(signed=False, label="UInt32"), "_bitfield": SimTypeInt(signed=False, label="UInt32")}, name="WHV_DOORBELL_MATCH_DATA", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition", "MatchData", "EventHandle"]),
        #
        'WHvUnregisterPartitionDoorbellEvent': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"GuestAddress": SimTypeLongLong(signed=False, label="UInt64"), "Value": SimTypeLongLong(signed=False, label="UInt64"), "Length": SimTypeInt(signed=False, label="UInt32"), "_bitfield": SimTypeInt(signed=False, label="UInt32")}, name="WHV_DOORBELL_MATCH_DATA", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Partition", "MatchData"]),
    }

lib.set_prototypes(prototypes)



#!/usr/bin/env python3
# pylint: disable=missing-class-docstring,disable=no-self-use
__package__ = __package__ or "tests.analyses"  # pylint:disable=redefined-builtin

import os
import unittest

import angr

from ..common import bin_location


test_location = os.path.join(bin_location, "tests")


class TestConstantpropagation(unittest.TestCase):
    def test_libc_x86(self):
        # disabling auto_load_libs increases the execution time.
        p = angr.Project(os.path.join(test_location, "i386", "libc-2.27-3ubuntu1.so.6"), auto_load_libs=True)
        dl_addr = p.loader.find_symbol("_dl_addr").rebased_addr
        cfg = p.analyses.CFGFast(regions=[(dl_addr, dl_addr + 4096)])
        func = cfg.functions["_dl_addr"]

        rtld_global_sym = p.loader.find_symbol("_rtld_global")
        assert rtld_global_sym is not None
        _rtld_global_addr = rtld_global_sym.rebased_addr

        base_addr = 0x998F000
        state = p.factory.blank_state()
        for addr in range(0, 0 + 0x1000, p.arch.bytes):
            state.memory.store(
                _rtld_global_addr + addr, base_addr + addr, size=p.arch.bytes, endness=p.arch.memory_endness
            )

        prop = p.analyses.Propagator(func=func, base_state=state)
        # import pprint
        # pprint.pprint(prop.replacements)
        assert len(prop.replacements) > 0

    def test_lwip_udpecho_bm(self):
        bin_path = os.path.join(test_location, "armel", "lwip_udpecho_bm.elf")
        p = angr.Project(bin_path, auto_load_libs=False)
        cfg = p.analyses.CFG(data_references=True)

        func = cfg.functions[0x23C9]
        state = p.factory.blank_state()
        prop = p.analyses.Propagator(func=func, base_state=state)

        assert len(prop.replacements) > 0


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
from unittest import main, TestCase

from angr.knowledge_plugins.key_definitions.heap_address import HeapAddress


class TestHeapAddress(TestCase):
    def test_expose_its_value_as_a_property(self):
        address = HeapAddress(0x42)
        self.assertEqual(address.value, 0x42)

    def test_add_fails_if_value_added_is_not_int(self):
        address = HeapAddress(0x0)
        self.assertRaises(TypeError, address + 0x8)

    def test_add_increase_the_heap_address_value_by_the_right_amount(self):
        address = HeapAddress(0x0)
        new_address = address + 0x8
        self.assertEqual(new_address.value, 0x8)

    def test_add_is_commutative(self):
        address = HeapAddress(0x0)
        self.assertEqual(address + 0x8, 0x8 + address)


if __name__ == "__main__":
    main()



import logging

from ....simos import SimWindows
from .resolver import IndirectJumpResolver

l = logging.getLogger(name=__name__)


class X86PeIatResolver(IndirectJumpResolver):
    """
    A timeless indirect jump resolver for IAT in x86 PEs.
    """

    def __init__(self, project):
        super().__init__(project, timeless=True)

    def filter(self, cfg, addr, func_addr, block, jumpkind):
        if not isinstance(self.project.simos, SimWindows):
            return False
        if jumpkind != "Ijk_Call":
            return False

        opnd = self.project.factory.block(addr).capstone.insns[-1].insn.operands[0]
        # Must be of the form: call ds:0xABCD
        if opnd.mem and opnd.mem.disp and not opnd.mem.base and not opnd.mem.index:
            return True
        return False

    def resolve(
        self, cfg, addr, func_addr, block, jumpkind, func_graph_complete: bool = True, **kwargs
    ):  # pylint:disable=unused-argument
        slot = self.project.factory.block(addr).capstone.insns[-1].insn.disp
        target = cfg._fast_memory_load_pointer(slot)
        if target is None:
            l.warning("Address %#x does not appear to be mapped", slot)
            return False, []

        if not self.project.is_hooked(target):
            return False, []

        dest = self.project.hooked_by(target)
        l.debug("Resolved target to %s", dest.display_name)
        return True, [target]



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("hrtfapo.dll")
prototypes = \
    {
        #
        'CreateHrtfApo': SimTypeFunction([SimTypePointer(SimStruct({"distanceDecay": SimTypePointer(SimStruct({"type": SimTypeInt(signed=False, label="HrtfDistanceDecayType"), "maxGain": SimTypeFloat(size=32), "minGain": SimTypeFloat(size=32), "unityGainDistance": SimTypeFloat(size=32), "cutoffDistance": SimTypeFloat(size=32)}, name="HrtfDistanceDecay", pack=False, align=None), offset=0), "directivity": SimTypePointer(SimStruct({"type": SimTypeInt(signed=False, label="HrtfDirectivityType"), "scaling": SimTypeFloat(size=32)}, name="HrtfDirectivity", pack=False, align=None), offset=0)}, name="HrtfApoInit", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="IXAPO"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["init", "xApo"]),
    }

lib.set_prototypes(prototypes)
